// Fast multiplication (Karatsuba)
#include "types.h"
#include "large_math.h"
#include "fast_mul.h"
#include "random.h"
#include <stdio.h>

// Optimized for max. speed on PC (i7, P4, 32bit CPU)
// multiplications of 512bit x 512bit (and smaller) are done directly, 
// by ABC book algorithm
//

static const u32 use_slow_multiply_order = 4; // 2^4 = 16 words = 512 bits
static const u32 use_slow_multiply_size_x32 = ((u32)1)<< use_slow_multiply_order; 


/*
u64 karatsuba_test(u32 x, u32 y)
{
u16 a = (u16)(x >> 16);
u16 b = (u16)(x & 0xFFFF);
u16 c = (u16)(y >> 16);
u16 d = (u16)(y & 0xFFFF);

u32 z2 = a*(u32)c; // 32 bit
u32 z0 = b*(u32)d; // 32 bit


//u32 xx = a + (u32)b; // 17 bit
//u32 yy = c + (u32)d; // 17 bit
//u64 z1 = xx*(u64)yy; // 34 bit


u32 xx = a + (u32)b; // 17 bit
u32 cx = xx >> 16;
xx &= 0xFFFF;

u32 yy = c + (u32)d; // 17 bit
u32 cy = yy >> 16;
yy &= 0xFFFF;

u64 z1 = xx*(u64)yy; // 32 bit

if(cx) z1 += ((u64)yy) << 16;

if(cy) z1 += ((u64)xx) << 16;

if((cx)&&(cy)) z1 += ((u64)0x100000000);

z1 = z1 - z2 - z0;

u64 result = (((u64)z2)<<32) + (z1<<16) + z0;

return result;
}
*/

//-----------

FAST_MUL_STAGE::FAST_MUL_STAGE(FAST_MUL_STAGE *lower_stage, u32 len)
{
low_stage = lower_stage;
FAST_MUL_STAGE::len = len;

z0 = new u32[1 + len];
z1 = new u32[1 + len];
z2 = new u32[1 + len];
xx = new u32[1 + len/2];
yy = new u32[1 + len/2];

z0[len] = 0;
z2[len] = 0;
xx[len/2] = 0;
yy[len/2] = 0;
}

FAST_MUL_STAGE::~FAST_MUL_STAGE()
{
FAST_MUL_STAGE::reset();    
delete[] yy;
delete[] xx;    
delete[] z2;
delete[] z1;
delete[] z0;
}
//---------
void FAST_MUL_STAGE::reset()
{
cheap_random_number(z0, 1 + len);
cheap_random_number(z1, 1 + len);
cheap_random_number(z2, 1 + len);
cheap_random_number(xx, 1 + len/2);
cheap_random_number(yy, 1 + len/2);

z0[len] = 0;
z2[len] = 0;
xx[len/2] = 0;
yy[len/2] = 0;
}

void FAST_MUL_STAGE::fast_mul(u32 *result, u32 const *x, u32 const *y)
{
if(low_stage == NULL) // compute directly
 {
 /*   // split to basic multiply - there is no point as splitting below 256 bit makes process slower due to overheads
 u64 tmp = x[0]*(u64)y[0];
 result[0] = (u32)(tmp & 0xFFFFFFFF);
 result[1] = (u32)(tmp >> 32);
 */
 if(len == 16) multi_mul_512(result, x, y); // about 50% faster
 else multi_mul(result, x, y, len);
 }
else // split to lower stages
 {
 low_stage->fast_mul(z0, x, y);                  //  z0 = B x D (low x low)
 low_stage->fast_mul(z2, x + len/2, y + len/2);  //  z2 = A x C (high x high)

 u32 cx = multi_add(xx, x, x + len/2, len/2); // xx = A + B
 u32 cy = multi_add(yy, y, y + len/2, len/2); // yy = C + D

 z1[len] = 0;
 
 low_stage->fast_mul(z1, xx, yy);  // z1 = low(xx) x low(yy) 
 
 if(cx) multi_add(z1 + len/2, z1 + len/2, yy, 1 + len/2); 

 if(cy) multi_add(z1 + len/2, z1 + len/2, xx, 1 + len/2); 
 
 if((cx) && (cy)) z1[len]++;

 multi_sub(z1, z1, z2, len + 1);
 multi_sub(z1, z1, z0, len + 1);

 copy(result, z0, len);
 copy(result + len, z2, len);

 if(multi_add(result + len/2, result + len/2, z1, len + 1)) multi_inc(result + len + len/2, len/2); 
 }
}
//------------------------------------------

FAST_MUL::FAST_MUL(u32 size_x32)
{
len = size_x32;

// Calculate number of stages

number_of_stages = 1;
u32 tmp = 1;


while(tmp < size_x32)
 {
 tmp <<= 1;
 number_of_stages++;
 }


// Allocate stages

fast_mul_stage = new FAST_MUL_STAGE *[number_of_stages];

FAST_MUL_STAGE *lower_stage = NULL;
tmp = 1;

for(u32 ci = 0; ci < number_of_stages; ci++)
 {
 if(tmp <= use_slow_multiply_size_x32) lower_stage = NULL;

 fast_mul_stage[ci] = lower_stage = new FAST_MUL_STAGE(lower_stage, tmp);
 tmp <<= 1;
 }
}

//-----------------------------------------------

FAST_MUL::~FAST_MUL()
{
for(u32 ci = number_of_stages; ci > 0; ci--)
 {
 delete fast_mul_stage[ci-1];
 }

delete[] fast_mul_stage;
}

//--------- Actual operation ------------
// Note: operands should not overlap with result!
//
void FAST_MUL::fast_mul(u32 *result, u32 const *x, u32 const *y)
{
fast_mul_stage[number_of_stages-1]->fast_mul(result, x, y);
}

//-------- Clean up all stages -------------
//

void FAST_MUL::reset(void)
{
for(u32 ci = 0; ci < number_of_stages; ci++)
 {
 fast_mul_stage[ci]->reset();
 }
}


//=================== <END> ==============


