// Hash functions 

#include "types.h"
#include "xtea.h"
#include "idea.h"
#include "hash_func.h"
#include "serialize.h"
#include "el_dumbo.h"


//------------------------------------------------------------
// Generate 64-bit hash of arbitrary data length
// using XTEA
// Initial value of hash must be initialized 
//
void Hash_Block_X_64(u32 *hash, u8 const* data, u32 data_length_bytes)
{
u32 data_input[2], lock[2];
u32 key[4]; 
u32 tweak = 0;

#define TWEAK_PRNG tweak = (((tweak << 5) - tweak) ^ 0x6A551EE9)

// Chaining mode with random tweak value
   
while(data_length_bytes)
 {
 data_length_bytes = fetch_dwords(data_input, 2, data, data_length_bytes);
 data += 2*sizeof(u32);

 lock[0] = data_input[0];
 lock[1] = data_input[1];
 key[0] = hash[0];
 key[1] = hash[1]; 
 key[2] = TWEAK_PRNG; 
 key[3] = TWEAK_PRNG;

 XTEA_Encode(data_input, key);

 hash[0] = data_input[0] ^ lock[0];
 hash[1] = data_input[1] ^ lock[1];
 }

#undef TWEAK_PRNG
}


//--------------------

COLLAPSE::COLLAPSE(u32 hash_size_x32)
{
size_x32 = hash_size_x32;

size_x128 = size_x32 >> 2;
if(size_x32 & 0x03) size_x128++;

data_input = new u64[2*size_x128];
hash_input = new u64[4*size_x128];
lock_bar   = new u64[2*size_x128];
el_dumbo_tmp = new u32[4*size_x128];
sc128 = new SC128_CIPHER(size_x128);


data_byte_cnt = 0;
tweak = 0x0123456789ABCDEF;

COLLAPSE::reset();
}

COLLAPSE::~COLLAPSE()
{
COLLAPSE::reset();

delete sc128;
delete[] el_dumbo_tmp;
delete[] lock_bar;
delete[] hash_input;
delete[] data_input;
}

void COLLAPSE::reset()
{
for(u32 i = 0; i < 2*size_x128; i++)
 {
 data_input[i] = COLLAPSE::prng();
 lock_bar[i]   = COLLAPSE::prng();
 hash_input[2*i + 0] = COLLAPSE::prng();
 hash_input[2*i + 1] = COLLAPSE::prng();
 }

for(u32 i = 0; i < 4*size_x128; i++) el_dumbo_tmp[i] = (u32)COLLAPSE::prng();

data_byte_cnt = 0;
}

void COLLAPSE::encrypt(u64 *data, u64 const* key)
{
El_Dumbo_2_Encode(XTEA_Encode,  (u32 *)data, el_dumbo_tmp, (u32 *)key, 2*size_x128);
sc128->Encode(data, key);
}

void COLLAPSE::decrypt(u64 *data, u64 const* key)
{
sc128->Decode(data, key);
El_Dumbo_2_Decode(XTEA_Decode,  (u32 *)data, el_dumbo_tmp, (u32 *)key, 2*size_x128);
}


void COLLAPSE::hash_primitive()
{
data_byte_cnt = 0;

for(u32 i = 0; i < (2*size_x128); i++)  lock_bar[i] = data_input[i];

for(u32 i = 0; i < (2*size_x128); i++)  hash_input[i + 2*size_x128] = COLLAPSE::prng();

COLLAPSE::encrypt(data_input, hash_input);

for(u32 i = 0; i < (2*size_x128); i++)
 {
 hash_input[i] = data_input[i] ^ lock_bar[i];
 data_input[i] = 0;
 }
}
//---------------
// xor64 by Marsaglia
//
u64 COLLAPSE::prng()
{

tweak ^= tweak << 13;
tweak ^= tweak >> 7;
tweak ^= tweak << 17;

return tweak;
}


// Call for every byte of data

void COLLAPSE::process_hash(const u8 data)
{
data_input[data_byte_cnt/8] |= ((u64)data) << (8*(data_byte_cnt & 0x07));
data_byte_cnt++;

if(data_byte_cnt >= (size_x128 * 16)) COLLAPSE::hash_primitive();
}

void COLLAPSE::complete_hash(u32 *output)
{
// Complete data

if(data_byte_cnt) COLLAPSE::hash_primitive();

// Output transform

for(u32 i = 0; i < (2*size_x128); i++) data_input[i] = 0;

COLLAPSE::hash_primitive();

// Output

for(u32 i = 0; i < size_x32; i++) output[i] =(u32)(0xFFFFFFFF & (hash_input[i/2] >> ((i & 1) * 32)));

// Cleanup

COLLAPSE::reset();
}

void COLLAPSE::init_hash(const u32 data_length_bytes, const u32 iv)
{
data_byte_cnt = 0;
tweak = 0x0123456789ABCDEF;

for(u32 i = 0; i < (2*size_x128); i++)
 {
 data_input[i] = 0;
 hash_input[i] = COLLAPSE::prng(); // start with nonzero constant
 }

// Include length and IV

u64 tmp = iv;
tmp = (tmp << 32) | data_length_bytes;

hash_input[0] ^= tmp;

// Compute first block

COLLAPSE::hash_primitive();

}

//------------

void COLLAPSE::process_hash(void const *data, u32 data_block_size_bytes)
{

u8 *ptr = (u8*)data;

while(data_block_size_bytes)
 {
 data_block_size_bytes--;
 COLLAPSE::process_hash(*ptr);
 ptr++;
 }
}

void COLLAPSE::hash(u32 *hash, void const *data, u32 data_block_size_bytes, const u32 iv)
{
COLLAPSE::init_hash(data_block_size_bytes, iv);
COLLAPSE::process_hash(data, data_block_size_bytes);
COLLAPSE::complete_hash(hash);
}




//-----------------------
// Compute hash of DH parameters
//

DH_HASH::DH_HASH(u32 size_x32)
{
collapse = new COLLAPSE(size_x32);
}

DH_HASH::~DH_HASH()
{
DH_HASH::reset();
delete collapse;
}


void DH_HASH::reset()
{
collapse->reset();
}

void DH_HASH::hash(u32 *hash, u32 const* g, u32 const* y, u32 const* p, const u32 dh_length_x32)
{
collapse->init_hash(3*dh_length_x32*sizeof(u32));
collapse->process_hash(g, dh_length_x32*sizeof(u32));
collapse->process_hash(y, dh_length_x32*sizeof(u32));
collapse->process_hash(p, dh_length_x32*sizeof(u32));
collapse->complete_hash(hash);
}

void DH_HASH::hash(u32 *hash, u32 const* g, u32 const *p, u32 const* gxp, u32 const* gyp, const u32 dh_length_x32)
{
collapse->init_hash(4*dh_length_x32*sizeof(u32));
collapse->process_hash(g, dh_length_x32*sizeof(u32));
collapse->process_hash(p, dh_length_x32*sizeof(u32));
collapse->process_hash(gxp, dh_length_x32*sizeof(u32));
collapse->process_hash(gyp, dh_length_x32*sizeof(u32));
collapse->complete_hash(hash);
}

#if 0
int collapse_test()
{
// 512 bit hash over 2kb of data ~ 2.7ms
//
const u32 hash_size_x32 = 16;
COLLAPSE cs(hash_size_x32);

/*
const u32 len_x64 = hash_size_x32/2; // 512 bit
u64 x[len_x64];
u64 k[2*len_x64];


for(u32 i = 0; i < len_x64; i++) x[i] = i + 1;
for(u32 i = 0; i < 2*len_x64; i++) k[i] = i;

printf("\n Encrypt/Decrypt:");
for(u32 i = 0; i < len_x64; i++) printf("\n 0x%016llx", x[i]);

cs.encrypt(x, k);

printf("\n");
for(u32 i = 0; i < len_x64; i++) printf("\n 0x%016llx", x[i]);

cs.decrypt(x, k);

printf("\n");
for(u32 i = 0; i < len_x64; i++) printf("\n 0x%016llx", x[i]);
*/
const u32 data_len = 2048;
u8 *data = new u8[data_len];
for(u32 i = 0; i < data_len; i++) data[i] = (u8)i;

data[2047] = 1;

u32 hash[hash_size_x32];

cs.hash(hash, data, data_len, 0x12345);

//for(u32 i = 0; i < 10000; i++) cs.hash(hash, data, data_len, 0x12345);

printf("\n");
for(u32 i = 0; i < hash_size_x32; i++) printf("\n 0x%08x", hash[i]);

return 1;
}
#endif


//===================== <END> =====================
