#include "types.h"
#include "keystream.h"
#include "sc128_lite.h"
#include "idea.h"
#include <string.h>

//------------------------------
// Xorshift 128 bit LSFR (Marsaglia)
//
inline static void keystream_xor128a(u32 *seed)
{
const u32 a = 15;
const u32 b = 14;
const u32 c = 1;

u32 cnt = 4; // full update of all 128 bits
while(cnt--)
 {
 u32 t = seed[0];  // t = x ^ (x << a)
 t ^= t << a;
 seed[0] = seed[1]; // x = y
 t ^= t >> b;
 seed[1] = seed[2]; // y = z
 u32 w = seed[2] = seed[3]; // z = w
 seed[3] = (w ^ t) ^ (w >> c);
 }
}

// Platform independent

inline static void keystream_32_to_64(u64 *x, u32 const *state)
{
x[0] = state[0] | (((u64)state[1])<<32);
x[1] = state[2] | (((u64)state[3])<<32);
x[2] = state[4] | (((u64)state[5])<<32);
x[3] = state[6] | (((u64)state[7])<<32);
x[4] = state[8] | (((u64)state[9])<<32);
x[5] = state[10] | (((u64)state[11])<<32);
}


inline static void keystream_64_to_32(u32 *state, u64 *x)
{
// No need to copy first 4 words, but for completeness
state[0] = (u32)(x[0] & 0xFFFFFFFF);
state[1] = (u32)(x[0] >> 32);
x[0] = 0;

state[2] = (u32)(x[1] & 0xFFFFFFFF);
state[3] = (u32)(x[1] >> 32);
x[1] = 0;

//--------
state[4] = (u32)(x[2] & 0xFFFFFFFF);
state[5] = (u32)(x[2] >> 32);
x[2] = 0;

state[6] = (u32)(x[3] & 0xFFFFFFFF);
state[7] = (u32)(x[3] >> 32);
x[3] = 0;


state[8] = (u32)(x[4] & 0xFFFFFFFF);
state[9] = (u32)(x[4] >> 32);
x[4] = 0;

state[10] = (u32)(x[5] & 0xFFFFFFFF);
state[11] = (u32)(x[5] >> 32);
x[5] = 0;
}



//----------------------
// Hash function
// 128 bit input
// 256 bit state
//
inline static void keystream_hash_256(u32 *state)
{
u64 x[keystream_state_size_x64];
u64 lock_bar[4];

keystream_32_to_64(x, state);

lock_bar[0] = x[2];
lock_bar[1] = x[3];
lock_bar[2] = x[4];
lock_bar[3] = x[5];

sc128_lite_128_256_encode(x + 2, x);

x[2] ^= lock_bar[0];
lock_bar[0] = 0;
x[3] ^= lock_bar[1];
lock_bar[1] = 0;
x[4] ^= lock_bar[2];
lock_bar[2] = 0;
x[5] ^= lock_bar[3];
lock_bar[3] = 0;

keystream_64_to_32(state, x);
}
//-----------------------------
// Generate 16 bytes of cipher gamma
//
inline static void keystream_blockcipher(u8 *byte_buffer, u32 const *state)
{
u32 tmp_buff[4];

tmp_buff[0] = state[0];
tmp_buff[1] = state[1];
tmp_buff[2] = state[2];
tmp_buff[3] = state[3];

IDEA_Encode(tmp_buff, state + 4, 2);

// 16 bytes

for(u32 ci = 0; ci < 4; ci++)
 {
 byte_buffer[ci + 0]  = (u8)(tmp_buff[0] & 0xFF);
 tmp_buff[0] >>= 8;

 byte_buffer[ci + 4]  = (u8)(tmp_buff[1] & 0xFF);
 tmp_buff[1] >>= 8;

 byte_buffer[ci + 8]  = (u8)(tmp_buff[2] & 0xFF);
 tmp_buff[2] >>= 8;

 byte_buffer[ci + 12] = (u8)(tmp_buff[3] & 0xFF);
 tmp_buff[3] >>= 8;
 }
}

//---------------------------------------------
// Generate output u8[16] from state u32[12]
//
void keystream_generator(u8 *output, u32 *state)
{
keystream_xor128a(state);
keystream_hash_256(state);
keystream_blockcipher(output, state);
}
//------------

void keystream_generator_init(u32 *state, u32 const *key)
{
for(u32 i = 0; i <  keystream_state_size_x32; i++) state[i] = key[i];

// Make sure LSFR is not set to all zero state

if(!(state[0] | state[1] | state[2] | state[3])) state[0] = 1;
}
//----------------

KEYSTREAM_GENERATOR::KEYSTREAM_GENERATOR()
{
KEYSTREAM_GENERATOR::reset();
}

KEYSTREAM_GENERATOR::~KEYSTREAM_GENERATOR()
{
KEYSTREAM_GENERATOR::reset();
}

void KEYSTREAM_GENERATOR::reset()
{
cntr = 0;
skip_cntr = 0;
byte_buffer_idx = 0;
memset(byte_buffer,0, sizeof(byte_buffer));
memset(state, 0, sizeof(state));
}

void KEYSTREAM_GENERATOR::init_key(u32 const* key)
{
keystream_generator_init(state, key);
cntr = 0;
skip_cntr = 0;
byte_buffer_idx = 0;
}

u8 KEYSTREAM_GENERATOR::get_keystream_byte()
{

if(!byte_buffer_idx) // time to generate next 16 bytes
 {
 keystream_generator(byte_buffer, state);
 }

u8 ret_value = byte_buffer[byte_buffer_idx & 0x0F]; // Could be = 16 in fast fwd mode
byte_buffer_idx = (byte_buffer_idx + 1) & 0x0F;

cntr++;

return ret_value;
}
//--------------------------------------
// Update state without generating keystream data
// (as if 16 bytes per iteration)
//
void KEYSTREAM_GENERATOR::fast_forward(u64 fwd_bytes)
{
if(!fwd_bytes) return; // Nothing to do

cntr += fwd_bytes;

u32 bytes_in_buffer = 0x0F & (16 - byte_buffer_idx); // 0...15

if(fwd_bytes >= bytes_in_buffer)
 {
 u64 diff_bytes = fwd_bytes - bytes_in_buffer;
 u64 fwd_cycles = 1 + (diff_bytes >> 4);

 byte_buffer_idx = (u32)(diff_bytes & 0x0F);
 if(!byte_buffer_idx) byte_buffer_idx = 16;

// Roll state forward by number of 16 byte blocks

 while(fwd_cycles)
  {
  fwd_cycles--;
  keystream_xor128a(state);
  keystream_hash_256(state);
  } // while

 // Generate buffer at last cycle

 keystream_blockcipher(byte_buffer, state);
 }
else // No need to update state, just move pointer
 {
 byte_buffer_idx += (u32)fwd_bytes;
 }
}

//------------------------------------------------


int KEYSTREAM_GENERATOR::counter(u64 x)
{
if(x < cntr) return 0; // Can't go back in time

/*
// Slow

while(x > cntr)
 {
 skip_cntr++;
 get_keystream_byte(); // Roll forward
 }
*/

if(x > cntr)
 {
 u64 diff = x - cntr;

 skip_cntr += diff;

 fast_forward(diff);
 }

return 1;
}

u64 KEYSTREAM_GENERATOR::counter()
{
return cntr;
}

u64 KEYSTREAM_GENERATOR::skip_counter()
{
return skip_cntr;
}
//------------------------------

int test_keystream()
{
KEYSTREAM_GENERATOR kg;

static const u32 key[keystream_state_size_x32] = { 0,1,2,3,4,5,6,7,8,9,10,11 };
kg.init_key(key);

static const u8 v0[32] =
{
0xBD,0x63,0x7B,0xDD,0x49,0x17,0xCD,0x44,
0x96,0xA6,0xE6,0x88,0x82,0x73,0x27,0x47,
0xBE,0xA8,0x71,0xCE,0xF9,0xCF,0xCD,0x13,
0xF3,0xBD,0x85,0x58,0x9E,0xD3,0x8B,0x60,
};

u8 test0[32];

for(u32 i = 0; i < 32; i++) test0[i] = kg.get_keystream_byte();

int ret_value = (memcmp(v0, test0, 32)) ?  0 : 1;

/*
for(u32 i = 0; i < 32; i++)
 {
 if(!(i&0x07)) printf("\n");
// printf("0x%02X,", test0[i]);
 printf("\n0x%02x = 0x%02x,", test0[i], v0[i]);
 }
*/

#if 0
// Test fast fwd
 {
 u32 seed[4] = { 1,2,3,4 };
 KEYSTREAM_GENERATOR kg0, kg1;
 kg0.init_key(key);
 kg1.init_key(key);


 for(u32 i = 0; i < 100000; i++)
  {
  keystream_xor128a(seed);
  u32 fwd = seed[0] & 0xFF;

  for(u32 j = 0; j < fwd; j++) kg0.get_keystream_byte();
  kg1.fast_forward(fwd);

  if(kg0.get_keystream_byte() != kg1.get_keystream_byte())
   {
   ret_value = 0;
   break;
   }
  }

 }
#endif

return ret_value;
}

//=========================== <END> ==============================

