#include "pkf.h"
#include <stdio.h>
#include <string.h>

//-------------------------------------------------
//  Password to key conversion function
//  Designed heavy on computation and memory
//
//----------- Using primitives from sc128 lite ------------------
//
// Optimized S-box
// abc=10,17,22 dev ~ 7e-002
//
static inline u64 sc128_lite_s_box(u64 x)
{
x = (x << 5) - x;
x ^= x >> 17;
x = (x << 22) | (x >> (64 - 22));
return x;
}

//------------------------------
// M-sequences of length = 64
// cyclically shifted by 0,1,2,3,4,5
//
static const u64 sc128_lite_key[6] =
{
0x7c10c53d1c96ecd5, //0
0xf8218a7a392dd9aa, //1
0xf04314f4725bb355, //2
0xe08629e8e4b766ab, //3
0xc10c53d1c96ecd57, //4
0x8218a7a392dd9aaf  //5
};

//-------------------------------
// Feistel primitive (6 rounds)
// Enough to get differential probability (~ 8 bit)
//
inline static void pkf_sc128_lite_feistel_64(u64 *x_0, u64 *x_1)
{
u64 x0 = *x_0;
u64 x1 = *x_1;

for(u32 i = 0; i <  6; i++)
 {
 u64 tmp = x0 + sc128_lite_s_box(x1 ^ sc128_lite_key[i]);
 x0 = x1;
 x1 = tmp;
 }

*x_0 = x0;
*x_1 = x1;
}
//---------------------
// State update function
// State is u64[3]
// key is u64[2]
//
static void pkf_sc128_round(u64 *state, u64 const *key)
{
state[0] ^= key[0];
state[1] ^= key[1];
pkf_sc128_lite_feistel_64(state + 0, state + 1);
pkf_sc128_lite_feistel_64(state + 1, state + 2);
pkf_sc128_lite_feistel_64(state + 2, state + 0);
}

//-----------
// result = x/y and round up
//
static u32 div_round_up(u32 x, u32 y)
{
u32 q = x / y;
u32 r = x % y;
if(r) q++;
return q;
}

void PKF::reset()
{
state[0] = 0;
state[1] = 1;
state[2] = 2;
for(u32 i = 0; i < mem_size_x32; i++) mem[i] = i;
}

PKF::PKF(u32 size_x32, u32 rounds)
{
if(!rounds) rounds = pkf_default_rounds; // just in case

PKF::rounds = rounds;

if(size_x32 > pkf_max_mem_size_x32) size_x32 = pkf_max_mem_size_x32;
else if(size_x32 < pkf_min_mem_size_x32) size_x32 = pkf_min_mem_size_x32;

// Find power of 2 more or equal to memory size

mem_size_x32 = 1;
while(size_x32 > mem_size_x32) mem_size_x32 <<= 1;

mem = new u32[mem_size_x32];
mem_idx_mask = mem_size_x32 - 1;

PKF::reset();
}

PKF::~PKF()
{
delete[] mem;
}


void PKF::get_key(void *key, const u32 key_size_bytes,
                  void const *salt, u32 salt_size_bytes,
                  void const *passwd, u32 passwd_size_bytes)
{
u64 k[2];
u32 n;
u32 wr_idx     = mem_size_x32 - 1;  // last address
u32 rd_idx     = mem_size_x32/2;    // in the middle
u32 cnt        = 0;
u32 passwd_idx = 0;
u32 salt_idx   = 0;
u32 key_idx    = 0;

if((key == NULL) || (key_size_bytes == 0)) return;
 

// Zero key initially

for(u32 i = 0; i < key_size_bytes; i++) ((u8*)key)[i] = 0;


// Check password and salt for validity

u8 foo = 0;

if((salt == NULL) || (salt_size_bytes == 0))
 {
 salt = &foo;
 salt_size_bytes = 1;
 }

if((passwd == NULL) || (passwd_size_bytes == 0))
 {
 passwd = &foo;
 passwd_size_bytes = 1;
 }


u32 r = rounds;


// Make sure there is enough rounds to utilize password and salt and initialize key

{
u32 iter = r * mem_size_x32; // total iterations
u32 passes = (8*iter)/key_size_bytes;

if(passes < pkf_min_passes)
 {
 iter = div_round_up(pkf_min_passes * key_size_bytes, 8);
 r = div_round_up(iter, mem_size_x32);
 }

passes = (2*iter)/passwd_size_bytes;
if(passes < pkf_min_passes)
 {
 iter = div_round_up(pkf_min_passes * passwd_size_bytes, 2);
 r = div_round_up(iter, mem_size_x32);
 }

passes = (2*iter)/salt_size_bytes;
if(passes < pkf_min_passes)
 {
 iter = div_round_up(pkf_min_passes * salt_size_bytes, 2);
 r = div_round_up(iter, mem_size_x32);
 }
}

//----------- Main loop ----------------

while(r--) // rounds through memory
 {
 for(u32 j = 0; j < mem_size_x32; j++)
  {
  // Read memory

  k[0]  =  mem[j];
  k[0]  = (k[0] << 32) | mem[rd_idx];

  
  // Cyclically read +2 bytes of password and +2 bytes of salt (total 4 bytes)

  k[1] = 0;
  n = 2;
  while(n--)
   {
   k[1] = (k[1] << 8) | ((u8 const*)passwd)[passwd_idx];
   passwd_idx++;
   if(passwd_idx >= passwd_size_bytes) passwd_idx = 0;

   k[1] =  (k[1] << 8) | ((u8 const*)salt)[salt_idx];
   salt_idx++;
   if(salt_idx >= salt_size_bytes) salt_idx = 0;
   }

  // Counter

  k[1] = (k[1] << 32) | cnt;
  cnt++;

  // Round function

  pkf_sc128_round(state, k);

  // Write memory

  mem[j] = (u32)(state[0] & 0xFFFFFFFF);
  mem[wr_idx] = (u32)(state[0] >> 32);

  // Update indexes

  rd_idx = (u32)(state[1] & mem_idx_mask);
  wr_idx = (u32)((state[1] >> 32) &  mem_idx_mask);

  
  // Update +8 bytes of key

  u64 tmp = state[2];
  n = 8;

  while(n--)
   {
   ((u8 *)key)[key_idx] ^= (u8)(tmp & 0xFF);
   tmp >>= 8;
   key_idx++;
   if(key_idx >= key_size_bytes) key_idx = 0;
   }
  } // j
 } // rounds

// Cleanup
PKF::reset();
}
//------------------------ Test -------------

int pkf_test()
{
PKF pkf;

const u32 key_len = 10;
const u8 key_test[key_len] = { 0x48, 0x79, 0x32, 0x8F, 0x9D, 0xFC, 0x1A, 0x80, 0xB9, 0x48 };
const char password[] = "012345";
const char salt[] = "6789ABCDEF";
u8 key[key_len];


pkf.get_key(key, key_len, password, strlen(password), salt, strlen(salt));

//printf("\nKey:");
//for(u32 i = 0; i < key_len; i++) printf(" 0x%02X,", key[i]);

return (memcmp(key, key_test, key_len)) ?  0 : 1;
}


//======================= <END> ===========================

