// Random numbers 

#include "types.h"
#include "xtea.h"
#include "random.h"
#include "system_random.h"
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <tchar.h>
#include "system_critical_section.h"
#include "el_dumbo.h"
#include "hash_func.h"
#include "keystream.h"

static SYSTEM_CRITICAL_SECTION main_rng_critical_section; // Must be static

// Main seed is 1 KB size
// There is two buffers (1 + 1Kb); entropy is collected into one, the other is used for random numbers
//
static const u32 entropy_mix_rounds = 4; // 4 rounds: more then 128 bit strong
static const u32 seed_size_u32 = 256;  // do not change
static u32 main_seed[2*seed_size_u32]; // main seed
static u32 cheap_seed[2]; // To generate junk
static u32 entropy_byte_cnt = 0;


//--------------------------------------
// Initialize seed

#ifndef _UNICODE

static char seed_path[MAX_PATH + 1] = "seed.bin";

void initialize_seed(const char *path)
{
if(path != NULL) strncpy(seed_path, path, sizeof(seed_path));


u8 buff[seed_size_u32*sizeof(u32)];

FILE *file = fopen(seed_path,"rb");

if(file != NULL)
 {
 fread(buff, sizeof(buff), 1, file); 
 fclose(file);
 collect_entropy(buff, sizeof(buff));
 }

// Wash seed

collect_entropy();

file = fopen(seed_path,"wb");
fwrite(main_seed, seed_size_u32*sizeof(u32), 1, file);
fclose(file);

// Initialize cheap seed
 
time_t t;
time(&t);
Hash_Block_X_64(cheap_seed, (u8 *)&t, sizeof(time_t));

// Cleanup

cheap_random_number(buff, seed_size_u32); 
reset_stack();
}
#else

static TCHAR seed_path[MAX_PATH + 1] = _T("seed.bin");

void initialize_seed(const TCHAR *path)
{
if(path != NULL) _tcsncpy(seed_path, path, MAX_PATH + 1);


u8 buff[seed_size_u32*sizeof(u32)];

FILE *file;

if(!_tfopen_s(&file,seed_path,_T("rb")))
 {
 fread(buff, sizeof(buff), 1, file); 
 fclose(file);
 collect_entropy(buff, sizeof(buff));
 }

// Wash seed

collect_entropy();

if(!_tfopen_s(&file,seed_path,_T("wb")))
 {
 fwrite(main_seed, seed_size_u32*sizeof(u32), 1, file);
 fclose(file);
 }

// Initialize cheap seed
 
time_t t;
time(&t);
Hash_Block_X_64(cheap_seed, (u8 *)&t, sizeof(time_t));

// Cleanup

cheap_random_number(buff, seed_size_u32); 
reset_stack();
}

#endif

//------------------------------------
// Uninvertible keyless function
//
static void mix_entropy(void)
{
u32 lock_bar[2*seed_size_u32];    

main_rng_critical_section.EnterCriticalSection();

for(u32 i = 0; i < 2*seed_size_u32; i++) lock_bar[i] = main_seed[i];

El_Dumbo_3_Encode(XTEA_Encode, lock_bar, seed_size_u32, entropy_mix_rounds); // size in 64 bit blocks

for(u32 i = 0; i < 2*seed_size_u32; i++) main_seed[i] ^= lock_bar[i];

main_rng_critical_section.LeaveCriticalSection();

cheap_random_number(lock_bar, 2*seed_size_u32); 
reset_stack();
}
//----------------------------
// Collect entropy
// When 1K is collected, seed and buffer are shuffled
//
//
void collect_entropy(void const *data, u32 data_size_bytes)
{


if((data == NULL) || (data_size_bytes == 0))
 {
 collect_entropy();
 return;
 }

main_rng_critical_section.EnterCriticalSection();

u8 *dst = (u8*)main_seed;
u8 *src = (u8*)data;

 
while(data_size_bytes--)
 {
 dst[entropy_byte_cnt] ^= *src++;
 entropy_byte_cnt++;

 if(entropy_byte_cnt >= seed_size_u32*sizeof(u32))  // 1/2 of buffer
  {   
  entropy_byte_cnt = 0;
  mix_entropy();
  }
 } // while


main_rng_critical_section.LeaveCriticalSection();
}


//-----------------------
// System provided entropy
// Note: system call can be easily captured
// So it should not be called in main RNG
//
void collect_entropy()
{

// Something for nothing
{
time_t t;
time(&t);
collect_entropy(&t, sizeof(time_t));
cheap_random_number((u8 *)&t, sizeof(time_t));
}


// Can we trust the system?

{
const u32 buff_size_bytes = 1024; // Draw 1K
u8 buff[buff_size_bytes];
Get_System_Provided_Entropy(buff, buff_size_bytes);
collect_entropy(buff, buff_size_bytes);
cheap_random_number(buff, sizeof(buff));
}

reset_stack();
}

//----
// u32[2] for legacy compatibility
//
void collect_entropy(u32 const* data)
{
collect_entropy(data, 2*sizeof(u32));
}



void reset_stack(void)
{
const u32 dummy_size = 256;    
u32 dummy[dummy_size];
cheap_random_number(dummy, dummy_size);
}

//----------------------------------------------------
// No need to protect cheap rng with critical sections
//

//---------------------------
// Simple reasonably good random numbers
// Period > 2^32
static u32 lcg64(u32 *state)
{
state[0] = 1664525*state[0] + 1013904223; // lcg from Numerical Recipies
state[1] = (state[1] << 7) | (state[1] >> (32 - 7));
state[1] += state[0]; // hash
return state[1];
}


u32 cheap_random_number()
{
return lcg64(cheap_seed);
}
 
void cheap_random_number(u32 *x, const u32 length_x32)
{
for(u32 i = 0; i < length_x32; i++) x[i] = lcg64(cheap_seed);
}

void cheap_random_number(u8 *x, const u32 size_of_x_bytes)
{
for(u32 i = 0; i < size_of_x_bytes; i++) x[i] = (u8)lcg64(cheap_seed);
}

//--------------------------------------------
// Generation of strong random numbers
//

void random_number(u32 *x, const u32 length_x32)
{
u32 tmp_state[2*seed_size_u32];
u32 tmp_x[2];

// Prehash
mix_entropy();

// Initialize RNG state

main_rng_critical_section.EnterCriticalSection();
for(u32 i = 0; i < (2*seed_size_u32); i++) tmp_state[i] = main_seed[i];
main_rng_critical_section.LeaveCriticalSection();

// Posthash
mix_entropy();

// Generate random numbers (counter mode)

tmp_x[1] = 0;

for(u32 i = 0; i < length_x32; i++)
 {
 tmp_x[0] = i;

 for(u32 j = 0; j < (2*seed_size_u32); j += 4)
  {
  XTEA_Encode(tmp_x, tmp_state + j);
  } // j
 x[i] = tmp_x[0]; // Use lower 32 bits; upper 32 bits are fed back so function can't be parallelized; also function is not bijective by counter
 } // i



//  Cleanup

cheap_random_number(tmp_x, 2);
cheap_random_number(tmp_state, 2*seed_size_u32);
reset_stack();
}

//==================== <END> ========================

