// Scalable cipher based on ARX Feistel primitive
// Operates on 64 bit words
// Block size = n * 128 bit
// Key size = 2 * block size
//
#include "sc128.h"
#include "sc128_lite.h"
#include "types.h"
#include <stdio.h>



static const u32 num_of_feistel_keys = 6;
static const u32 seed_size_u32 = 5;
static const u32 shft_value = 61; // simple number close to half block, to break symmetry of used sequences
static const u32 round_multiplier = 32; // x that many rounds times number of 128 bit blocks

static bool sc128_precalculated_keys_initialized = false;
static SC128_PRECALCULATED_KEYS sc128_precalculated_keys; // must be static; so to be initialized before main()


// slightly nonlinear prng (Marsaglia)
// xorshift + Weyl

static u32 sc128_rng(u32 *seed)
{
const u32 a = 15;
const u32 b = 14;
const u32 c = 1;

u32 t = seed[0];  // t = x ^ (x << a)
t ^= t << a;
seed[0] = seed[1]; // x = y
t ^= t >> b;
seed[1] = seed[2]; // y = z
u32 w = seed[2] = seed[3]; // z = w
seed[3] = (w ^ t) ^ (w >> c);
seed[4] += 0x9E3779B9; // 32 bit golden section

return (seed[3] ^ seed[4]);
}

//----------------
// Shift by +61 bit
//
static void sc128_shft_61(u64 *data, const u32 len_64)
{
const u32 shft = shft_value;
u64 tmp, data_current, data_next;

data_next = data[0];
tmp = data_next >> (64 - shft);
data_current = data_next << shft;

for(u32 i = 0; i < (len_64 - 1); i++)
 {
 data_next = data[i+1];
 data[i] = data_current | (data_next >> (64 - shft));
 data_current = data_next << shft;
 }

data[len_64 - 1] = data_current | tmp;
}
//----------------
// Shift by -61 bit
//
static void sc128_inv_shft_61(u64 *data, const u32 len_64)
{
const u32 shft = shft_value;
u64 tmp, data_current, data_next;

data_next = data[len_64 - 1];
tmp = data_next << (64 - shft);
data_current = data_next >> shft;

for(u32 i = (len_64 - 1); i > 0; i--)
 {
 data_next = data[i-1];
 data[i] = data_current | (data_next << (64 - shft));
 data_current = data_next >> shft;
 }

data[0] = data_current | tmp;
}


//---------------------------------------
// Differentially optimized parameters: 5,18,22 in 6-round feistel
//
static inline u64 sc128_s_box(u64 x)
{
x = (x << 5) + x;
x ^= x >> 18;
x = (x << 22) | (x >> (64 - 22));
return x;
}
//--------------------------------------------------------------------------------
// Constants made of M-sequences of 64 bit length with minimum running correlation
// to make every feistel round different even if keys are the same
// and to reduce rotation/shift symmetry
//
static const u64 sc128_ab[num_of_feistel_keys*2] =
 {
 0x7c10c53d1c96ecd5, 0x121c17e5467ad69b, // 01 (3)
 0x508b7eb8cec1e4a9, 0x93c1b98ebf68854a, // 35 (3)
 0x6b4d890e0bf2a33d, 0xa9fa0e12365ad798, // 14 (3)
 0x559bb49c5e51841f, 0x7a392dd9aaf8218a, // 20 (3)
 0x5e62a7e83848d96b, 0xa4f06e63afda2152, // 45 (4)
 0x2f28c20faacdda4e, 0x116fd719d83c952a  // 23 (4) shifted
 };



//-------------------------------
// Feistel primitive (6 rounds)
// Enough to get ~3.5e-2 differential probability (~ 9 bit)
//
static void sc128_feistel_64(u64 *x, u64 const* key)
{
u64 x0 = x[0];
u64 x1 = x[1];

for(u32 i = 0; i <  num_of_feistel_keys; i++) // 6 rounds, each round is unique, no (?) weak keys
 {
 u64 tmp = x0 + (sc128_ab[2*i + 1] ^ (sc128_ab[2*i + 0] + sc128_s_box(x1 ^ key[i])));
 x0 = x1;
 x1 = tmp;
 }

x[0] = x0;
x[1] = x1;
}

static void sc128_inv_feistel_64(u64 *x, u64 const* key)
{
u64 x1 = x[0];
u64 x0 = x[1];

u32 i = num_of_feistel_keys;
while(i--)
 {
 u64 tmp = x0 - (sc128_ab[2*i + 1] ^ (sc128_ab[2*i + 0] + sc128_s_box(x1 ^ key[i])));
 x0 = x1;
 x1 = tmp;
 }

x[0] = x1;
x[1] = x0;
}

//-----------------------------------------------
// Generate 6 random keys for feistel primitive
//
static void sc128_init_key(u64 *key, u32 *seed)
{
u64 x64[2];
u64 tmp64;

// Make random keys without obvious linear dependencies

for(u32 j = 0; j < num_of_feistel_keys; j++)
 {
 x64[0] = 0;
 x64[1] = 0;

 for(u32 i = 0; i < 16; i++)
  {
  tmp64 = sc128_rng(seed);
  tmp64 = (tmp64 << 32) | sc128_rng(seed);
  x64[0] ^= tmp64;

  tmp64 = sc128_rng(seed);
  tmp64 = (tmp64 << 32) | sc128_rng(seed);
  x64[1] ^= tmp64;

  sc128_lite_feistel_64(x64);
  } // i

 key[j] = x64[1]; // just one at time
 } // j
}

//-------------------
// Generate N feistel keys
//
static void sc128_init_keys(u64 *keys, u32 *seed, const u32 N)
{
for(u32 i = 0; i < N; i++)
 {
 sc128_init_key(keys, seed);
 keys += num_of_feistel_keys;
 }
}

//------------------------
// Round
//
static void sc128_round(u64 *x, u64 const *fptr, const u32 len_64)
{

for(u32 i = 0; i < len_64; i += 2)
 {
 sc128_feistel_64(x + i, fptr);
 fptr += num_of_feistel_keys;
 }
}


static void sc128_inv_round(u64 *x, u64 const *feistel_key, const u32 len_64)
{

for(u32 i = 0; i < len_64; i += 2)
 {
 sc128_inv_feistel_64(x + i, feistel_key);
 feistel_key += num_of_feistel_keys;
 }
}
//-------------------------------------------------------------------
// Apply key
//
static void sc128_apply_key(u64 *x, u64 const* k, const u32 data_len_64)
{
for(u32 i = 0; i < data_len_64; i++) x[i] ^= k[i];
}

// Entire cipher

static void sc128_encode(u64 *x, u64 const *key, u64 const *f_ptr, const u32 rounds, const u32 data_len_64)
{
u64 const *k0 = key;
u64 const *k1 = key + data_len_64;
u64 const f_inc = num_of_feistel_keys*data_len_64/2;

for(u32 i = 0; i < (rounds - 1); i++)
 {
 sc128_apply_key(x, k0, data_len_64);
 sc128_round(x, f_ptr, data_len_64);
 sc128_shft_61(x, data_len_64);
 f_ptr += f_inc;

 sc128_apply_key(x, k1, data_len_64);
 sc128_round(x, f_ptr, data_len_64);
 sc128_shft_61(x, data_len_64);
 f_ptr += f_inc;
 } // i

// + Last round

sc128_apply_key(x, k0, data_len_64);
sc128_round(x, f_ptr, data_len_64);
sc128_apply_key(x, k1, data_len_64);
}

// Entire decipher

static void sc128_decode(u64 *x, u64 const *key, u64 const *f_ptr, const u32 rounds, const u32 data_len_64)
{
u64 const *k0 = key;
u64 const *k1 = key + data_len_64;
u64 const f_inc = num_of_feistel_keys*data_len_64/2;

// + Last round

sc128_apply_key(x, k1, data_len_64);
sc128_inv_round(x, f_ptr, data_len_64);
sc128_apply_key(x, k0, data_len_64);

for(u32 i = 0; i < (rounds - 1); i++)
 {
 f_ptr -= f_inc;
 sc128_inv_shft_61(x, data_len_64);
 sc128_inv_round(x, f_ptr, data_len_64);
 sc128_apply_key(x, k1, data_len_64);

 f_ptr -= f_inc;
 sc128_inv_shft_61(x, data_len_64);
 sc128_inv_round(x, f_ptr, data_len_64);
 sc128_apply_key(x, k0, data_len_64);
 } // i

}

//------------------
SC128_PRECALCULATED_KEYS::SC128_PRECALCULATED_KEYS(u32 max_size_x128)
{
sc128_precalculated_size = max_size_x128;

u32 rounds = sc128_precalculated_size * round_multiplier;
u32 round_ops = sc128_precalculated_size *(2*rounds - 1);
u32 feistel_key_len_64 = num_of_feistel_keys * round_ops;

keys = new u64[feistel_key_len_64];

u32 seed[seed_size_u32] = { 0x12345678, 0x9ABCDEF0, 0xAA551EE1, 0xAAAA5555, 0 }; // Nothing in my head

sc128_init_keys(keys, seed, round_ops);

sc128_precalculated_keys_initialized = true; // global table is initialized
}

SC128_PRECALCULATED_KEYS::~SC128_PRECALCULATED_KEYS()
{
delete[] keys;
}

u64 const* SC128_PRECALCULATED_KEYS::precalculated_keys()
{
return keys;
}

u32 SC128_PRECALCULATED_KEYS::max_precalculated_size_x128()
{
return sc128_precalculated_size;
}


//------------------------------
// Width in 128 bit blocks
//
SC128_CIPHER::SC128_CIPHER(u32 width_128)
{
len_128 = width_128;     // block size

N_rounds = len_128 * round_multiplier; // one round is one use of the key, i.e. two runs through data block

u32 round_ops = len_128 *(2*N_rounds - 1);

feistel_key_len_64 = num_of_feistel_keys * round_ops;


if((sc128_precalculated_keys_initialized == true) && (len_128 <= sc128_precalculated_keys.max_precalculated_size_x128())) // already initialized?
 {
 local_precalculated_keys = NULL;
 feistel_key = sc128_precalculated_keys.precalculated_keys();
 }
else // create new keys
 {
 local_precalculated_keys = new SC128_PRECALCULATED_KEYS(len_128);
 feistel_key = local_precalculated_keys->precalculated_keys();
 }


SC128_CIPHER::Reset();
}

SC128_CIPHER::~SC128_CIPHER()
{
SC128_CIPHER::Reset();

if(local_precalculated_keys != NULL) delete local_precalculated_keys;
}

void SC128_CIPHER::Reset()
{
// Nothing to do
}

void SC128_CIPHER::Encode(u64 *x, u64 const *key)
{
sc128_encode(x, key, feistel_key, N_rounds, 2*len_128);
}

void SC128_CIPHER::Decode(u64 *x, u64 const *key)
{
sc128_decode(x, key, feistel_key + feistel_key_len_64 - num_of_feistel_keys*len_128, N_rounds, 2*len_128);
}
//=======================
#if 0
int SC128_Test()
{
const u32 len = 32;

SC128_CIPHER sc128(len);

u64 x[2*len];
u64 k[4*len];
for(u32 i = 0; i < 2*len; i++) x[i] = i;
for(u32 i = 0; i < 4*len; i++) k[i] = i;

sc128.Encode(x, k);

for(u32 i= 0; i < 2*len; i++) printf("\n %016llx,", x[i]);

sc128.Decode(x, k);

for(u32 i= 0; i < 2*len; i++) printf("\n %016llx,", x[i]);

/*
FILE *file = fopen("test.bin","wb");
u64 const *ptr = sc128_precalculated_keys.precalculated_keys();


u32 N_rounds = sc128_precalculated_keys.max_precalculated_size_x128() * round_multiplier; // one round is one use of the key, i.e. two runs through data block

u32 round_ops = sc128_precalculated_keys.max_precalculated_size_x128() *(2*N_rounds - 1);

u32 feistel_key_len_64 = num_of_feistel_keys * round_ops;


fwrite(ptr, feistel_key_len_64, 8, file);

fclose(file);
*/

return 1;
}
#endif

//============================ <END> =======================================

