// SC128 Lite cipher

#include "types.h"
#include <stdio.h>

static const u32 sc128_lite_shft = 61;
static const u32 num_of_feistel_keys = 6;

//
// Optimized S-box
// abc=10,17,22 dev ~ 7e-002
//
static inline u64 sc128_lite_s_box(u64 x)
{
x = (x << 5) - x;
x ^= x >> 17;
x = (x << 22) | (x >> (64 - 22));
return x;
}

//------------------------------
// M-sequences of length = 64
// cyclically shifted by 0,1,2,3,4,5
//
static const u64 sc128_lite_key[num_of_feistel_keys] =
{
0x7c10c53d1c96ecd5, //0
0xf8218a7a392dd9aa, //1
0xf04314f4725bb355, //2
0xe08629e8e4b766ab, //3
0xc10c53d1c96ecd57, //4
0x8218a7a392dd9aaf  //5
};

//-------------------------------
// Feistel primitive (6 rounds)
// Enough to get differential probability (~ 8 bit)
//
void sc128_lite_feistel_64(u64 *x)
{
u64 x0 = x[0];
u64 x1 = x[1];

for(u32 i = 0; i <  num_of_feistel_keys; i++)
 {
 u64 tmp = x0 + sc128_lite_s_box(x1 ^ sc128_lite_key[i]);
 x0 = x1;
 x1 = tmp;
 }

x[0] = x0;
x[1] = x1;
}

static void sc128_lite_inv_feistel_64(u64 *x)
{
u64 x1 = x[0];
u64 x0 = x[1];

u32 i = num_of_feistel_keys;
while(i--)
 {
 u64 tmp = x0 - sc128_lite_s_box(x1 ^ sc128_lite_key[i]);
 x0 = x1;
 x1 = tmp;
 }

x[0] = x1;
x[1] = x0;
}

// Two feistel primitives in parallel
// ~ 25 % faster then one by one
static void sc128_lite_dual_feistel_64(u64 *x)
{
u64 x0 = x[0];
u64 x1 = x[1];
u64 x2 = x[2];
u64 x3 = x[3];


for(u32 i = 0; i <  num_of_feistel_keys; i++)
 {
 u64 k = sc128_lite_key[i];
 u64 xx1 = x1 ^ k;
 u64 xx3 = x3 ^ k;
 u64 xxx1 = xx1 << 5;
 u64 xxx3 = xx3 << 5;
 xxx1 -= xx1;
 xxx3 -= xx3;
 xx1 = xxx1 >> 17;
 xx3 = xxx3 >> 17;
 xxx1 ^= xx1;
 xxx3 ^= xx3;
 xxx1 = (xxx1 << 22) | (xxx1 >> (64 - 22));
 xxx3 = (xxx3 << 22) | (xxx3 >> (64 - 22));
 xxx1 += x0;
 xxx3 += x2;

 x0 = x1;
 x1 = xxx1;
 x2 = x3;
 x3 = xxx3;
 }

x[0] = x0;
x[1] = x1;
x[2] = x2;
x[3] = x3;
}


inline static void sc128_lite_shft_61(u64 *x)
{
u32 shft = sc128_lite_shft;
u64 tmp = x[0] >> (64 - shft);
x[0] = (x[0] << shft) | (x[1] >> (64 - shft));
x[1] = (x[1] << shft) | tmp;
}

inline static void sc128_lite_inv_shft_61(u64 *x)
{
u32 shft = sc128_lite_shft;
u64 tmp = x[1] << (64 - shft);
x[1] = (x[1] >> shft) | (x[0] << (64 - shft));
x[0] = (x[0] >> shft) | tmp;
}


void sc128_lite_encode(u64 *x, u64 const *key, const u32 len_128)
{
u32 rounds = 32 * len_128;
u32 len = 2 * len_128;
u32 n = 0;

u64 xtmp[2];

for(u32 i = 0; i < rounds; i++)
 {

 for(u32 j = 0; j < (len - 1); j++)
  {
  x[j + 0] ^= key[2*j + 0];
  x[j + 1] ^= key[2*j + 1]  + n;
  n++;
  sc128_lite_feistel_64(x + j);
  sc128_lite_shft_61(x + j);
  }

 xtmp[0] = x[len-1];
 xtmp[1] = x[0];
 xtmp[0] ^= key[2*(len-1) + 0];
 xtmp[1] ^= key[2*(len-1) + 1] + n;
 n++;
 sc128_lite_feistel_64(xtmp);
 sc128_lite_shft_61(xtmp);
 x[len-1] = xtmp[0];
 x[0] = xtmp[1];
 } // rounds
}

void sc128_lite_decode(u64 *x, u64 const *key, const u32 len_128)
{
u32 rounds = 32 * len_128; // this is ~
u32 len = 2 * len_128;
u32 n = len*rounds;
u64 xtmp[2];

for(u32 i = 0; i < rounds; i++)
 {
 xtmp[0] = x[len-1];
 xtmp[1] = x[0];

 sc128_lite_inv_shft_61(xtmp);
 sc128_lite_inv_feistel_64(xtmp);

 n--;
 xtmp[0] ^= key[2*(len-1) + 0];
 xtmp[1] ^= key[2*(len-1) + 1] + n;

 x[len-1] = xtmp[0];
 x[0] = xtmp[1];

 u32 j = len - 1;
 while(j--)
  {
  sc128_lite_inv_shft_61(x + j);
  sc128_lite_inv_feistel_64(x + j);
  n--;
  x[j + 0] ^= key[2*j + 0];
  x[j + 1] ^= key[2*j + 1]  + n;
  }
 } // rounds
}
//-----------------------------------------
// Litest variant (block = 256, key = 128)
//
void sc128_lite_128_256_encode(u64 *x, u64 const *key)
{
u32 rounds = 36;
u32 n = 0;

while(rounds--)
 {
 x[0] ^= key[0];
 x[1] ^= key[1] + n;
 n++;

 x[2] ^= key[0];
 x[3] ^= key[1] + n;
 n++;

 // Slower
 //sc128_lite_feistel_64(x + 0);
 //sc128_lite_feistel_64(x + 2);

 // Faster by ~25%

 sc128_lite_dual_feistel_64(x);

 sc128_lite_shft_61(x + 1);
 } // rounds

// Last keying

x[0] ^= key[0];
x[1] ^= key[1] + n;
n++;

x[2] ^= key[0];
x[3] ^= key[1] + n;
}

void sc128_lite_128_256_decode(u64 *x, u64 const *key)
{
u32 rounds = 36;
u32 n = 2*rounds + 1;

x[2] ^= key[0];
x[3] ^= key[1] + n;

x[0] ^= key[0];
n--;
x[1] ^= key[1] + n;

while(rounds--)
 {
 sc128_lite_inv_shft_61(x + 1);
 sc128_lite_inv_feistel_64(x + 0);
 sc128_lite_inv_feistel_64(x + 2);

 x[2] ^= key[0];
 n--;
 x[3] ^= key[1] + n;

 x[0] ^= key[0];
 n--;
 x[1] ^= key[1] + n;
 } // rounds
}

//===================== <END> ============================

