#ifndef SIGNATURE_H_INCLUDED
#define SIGNATURE_H_INCLUDED

#include "types.h"
#include "fast_modulo.h"
#include "large_math.h"
#include "inverse.h"
#include "montgomery.h"
#include "pkf.h"
#include <stdio.h>
#include "hash_func.h"
#include "simple_test.h"

static const u32 signature_length_x32 = 128; // 4096 bit do not change
static const u32 fingerprint_length_x32 = 8; // 256 bit do not change
static const u32 user_info_max_length_bytes = 4*(signature_length_x32 - fingerprint_length_x32 - 8) - 2; // = 446 bytes; leave space for fingerprint and hash256, 
                                                                                                     // use two bytes for length
static const u32 certificate_length_bytes = 2016;

static const u32 salt_length_x32 = 8;
//
// data length in bits = 2 bit smaller then bit length of P
//


class ENCODED_SECRET_KEY 
{
public:
u32 key[signature_length_x32];
u32 hash;
u32 salt[salt_length_x32];
void reset();
};

void encode_secret_key(ENCODED_SECRET_KEY *esc, u32 const * secret_key, u8 const * password, const u32 password_length_bytes);
int decode_secret_key(u32 * secret_key, ENCODED_SECRET_KEY const* esc, u8 const * password, const u32 password_length_bytes); // Returns nonzero on success
int change_password(ENCODED_SECRET_KEY * esc,                                         // Returns nonzero on success
                    u8 const * old_password, const u32 old_password_length_bytes,
                    u8 const * new_password, const u32 new_password_length_bytes);

int check_password(ENCODED_SECRET_KEY const *esc, u8 const * password, const u32 password_length_bytes);

class SIGNATURE
{
public:    
u32 fingerprint[fingerprint_length_x32]; // 256 bit hash of certificate to read this signature
u32 r[signature_length_x32]; // part r
u32 s[signature_length_x32]; // part s
void reset();
};

class SERTIFICATE
{
public:
u16 user_info_length_bytes; // note:  this is 14-bit number, two msbs are zeroed
u8 user_info[user_info_max_length_bytes];  
u32 fingerprint[fingerprint_length_x32]; // 256 bit hash of certificate
u32 p[signature_length_x32]; // modulo
u32 g[signature_length_x32]; // generator
u32 y[signature_length_x32]; // open key
void reset();
};

// Our secret data 

class SECRET_DATA
{
public:
ENCODED_SECRET_KEY encoded_secret_key; // Password protected secret key for signing
SERTIFICATE certificate;     // Our certificate (not secret but stored together with secret key, so we could export it)
u8 data_is_initialized;
void reset();
};



class DIGITAL_SIGNATURE
{
u32 len;
u32 *q;
u32 *r_;
u32 *tmp; 
u32 *tmp1;
u32 *tmp2;
u32 *tmp_prod;

u32 *k;
u32 *g_;

u32 tmp_u8_len;
u8* tmp_u8; // For certificate fingerprint calculation

   
MONTGOMERY_POW_MOD *fmpm;
MULTI_DIV *multi_div;
INVERSE_MODULO *im;
TEST_GENERATOR_MODULO_P *tgmp;
DH_HASH *dh_hash;

int inverse_g_is_precalculated;
u32 precalculated_certificate_fingerprint[fingerprint_length_x32];


void calculate_fingerprint(u32 *fingerprint, SERTIFICATE const* sert); // 256-bit fingerprint u32[8]


public:
    DIGITAL_SIGNATURE(u32 length_x32 = signature_length_x32);
    ~DIGITAL_SIGNATURE();

    // Generate certificate and secret key using safe prime P, include optional user info

    void generate_keys(SERTIFICATE * sert, u32 *secret_key, u32 const *p, u8 const *userinfo = NULL, u16 userinfo_length_bytes = 0);
  
    void generate_keys(SECRET_DATA *secret_data, u8 const *password, u32 const password_length_bytes, 
                       u8 const *userinfo = NULL, u16 userinfo_length_bytes = 0);
  
    void include_user_info(SERTIFICATE * sert, u8 const *userinfo, u16 userinfo_length_bytes);

    // Decode signature using certificate
    void decode_signature(u32 *decoded_signature, SIGNATURE const* ds, SERTIFICATE const* sert);
    
    // Encode signature using certificate and secret key
    void encode_signature(SIGNATURE *ds, u32 *data, SERTIFICATE const* sert, u32 const* secret_key);

    // Returns nonzero if password match
    
    int encode_signature(SIGNATURE *ds, u32 *data, SECRET_DATA const *secret_data, u8 const *password, u32 const password_length_bytes);
    int encode_signature(SIGNATURE *ds, u32 *data, SERTIFICATE const* sert, ENCODED_SECRET_KEY const *esc, u8 const *password, const u32 password_length_bytes);

   
    int certificate_looks_ok(SERTIFICATE const* sert); // checks if certificate looks valid

    u32 serialize_certificate(u8 *bytes, SERTIFICATE const* sert); // returns length in bytes = 4*3*128 + 4*8 = 1568 bytes
    u32 deserialize_certificate(SERTIFICATE* sert, u8 const * bytes); // returns length in bytes = 4*3*128 + 4*8 = 1568 bytes


    u32 serialize_signature(u8 *bytes, SIGNATURE const *sign); // returns length in bytes = 4*2*128 = 1024 bytes
    u32 deserialize_signature(SIGNATURE *sign, u8 const *bytes);

    void reset();

};

void compose_signature_data(u32 *data, u32 const *dh_hash, SERTIFICATE const* sert);
void decompose_signature_data(u32 *dh_hash,
                              u32 *fingerprint, 
                              u8 *userinfo,
                              u16 *userinfo_length_bytes,
                               u32 const* data);

//------------------------------------------
// Sign anything with full length signature
// Returns nonzero on success

int calculate_signature(SIGNATURE *signature, SECRET_DATA  const *secret_data,
           u8 const *password, u32 const password_length_bytes,
            void const *data, u32 const data_size_bytes);

// Verify signature against certificate
// Returns error code

const int VERIFY_SIGNATURE_OK = 0;
const int VERIFY_SIGNATURE_BAD_SERTIFICATE   = 1; // bad certificate
const int VERIFY_SIGNATURE_WRONG_SERTIFICATE = 2; // signed with different certificate
const int VERIFY_SIGNATURE_DOES_NOT_MATCH    = 3; // signature does not match


int verify_signature(SIGNATURE const *signature, SERTIFICATE  const *certificate,
                     void const *data, u32 const data_size_bytes);

//----- Class for XEX encryption -------------- 

static const u32 xex_data_block_size_x32 = 2;
static const u32 xex_key_size_x32 = 10;

class XEX_ENCRYPT_DECRYPT
 {
 COLLAPSE *collapse;
 int was_initialized;
 u32 key[xex_key_size_x32]; // cntr[2] + key0[4] + key1[4]
 
 public: 
    XEX_ENCRYPT_DECRYPT();
    ~XEX_ENCRYPT_DECRYPT();
    void reset();
    int initialized();
    int Init(const u32 *salt,  SECRET_DATA const* secret_data, u8 const *password, const u32 password_length_bytes); // Returns nonzero on success
    void Encrypt(u32 *data, const u64 block_number); // Encrypt block of u32[2]
    void Decrypt(u32 *data, const u64 block_number); // Decrypt block of u32[2]
    void Encrypt(u32 *data, const u64 starting_block_number, const u64 number_of_blocks); // Encrypt block of u32[2]
    void Decrypt(u32 *data, const u64 starting_block_number, const u64 number_of_blocks); // Decrypt block of u32[2]
 
 };


#endif

